@@ -5,21 +5,50 @@
 
 """ A course class with some modification for users """
 
+import copy
 import gettext
+import re
+from typing import List
 from collections import OrderedDict
-from natsort import natsorted
 
-from inginious.common.courses import Course
 from inginious.common.tags import Tag
 from inginious.frontend.accessible_time import AccessibleTime
 from inginious.frontend.parsable_text import ParsableText
+from inginious.frontend.user_manager import UserInfo
+from inginious.frontend.task_dispensers.toc import TableOfContents
+
+def _migrate_from_v_0_6(content, task_list):
+    if 'task_dispenser' not in content:
+        content["task_dispenser"] = "toc"
+        if 'toc' in content:
+            content['dispenser_data'] = content["toc"]
+        else:
+            ordered_tasks = OrderedDict(sorted(list(task_list.items()),
+                                               key=lambda t: (int(t[1]._data.get('order', -1)), t[1].get_id())))
+            indexed_task_list = {taskid: rank for rank, taskid in enumerate(ordered_tasks.keys())}
+            content['dispenser_data'] = [{"id": "tasks-list", "title": _("List of exercises"),
+                                          "rank": 0, "tasks_list": indexed_task_list}]
 
 
-class WebAppCourse(Course):
+class Course(object):
     """ A course with some modification for users """
 
-    def __init__(self, courseid, content, course_fs, task_factory, hook_manager):
-        super(WebAppCourse, self).__init__(courseid, content, course_fs, task_factory, hook_manager)
+    def __init__(self, courseid, content, course_fs, task_factory, plugin_manager, task_dispensers):
+        self._id = courseid
+        self._content = content
+        self._fs = course_fs
+        self._task_factory = task_factory
+        self._plugin_manager = plugin_manager
+
+        self._translations = {}
+        translations_fs = self._fs.from_subfolder("$i18n")
+        if translations_fs.exists():
+            for f in translations_fs.list(folders=False, files=True, recursive=False):
+                lang = f[0:len(f) - 3]
+                if translations_fs.exists(lang + ".mo"):
+                    self._translations[lang] = gettext.GNUTranslations(translations_fs.get_fd(lang + ".mo"))
+                else:
+                    self._translations[lang] = gettext.NullTranslations()
 
         try:
             self._name = self._content['name']
@@ -29,6 +58,8 @@ class WebAppCourse(Course):
         if self._content.get('nofrontend', False):
             raise Exception("That course is not allowed to be displayed directly in the webapp")
 
+        _migrate_from_v_0_6(content, self._task_factory.get_all_tasks(self))
+
         try:
             self._admins = self._content.get('admins', [])
             self._tutors = self._content.get('tutors', [])
@@ -41,13 +72,18 @@ class WebAppCourse(Course):
                 raise Exception("Course has an invalid value for registration_ac: " + self.get_id())
             self._registration_ac_list = self._content.get('registration_ac_list', [])
             self._groups_student_choice = self._content.get("groups_student_choice", False)
-            self._use_classrooms = self._content.get('use_classrooms', True)
             self._allow_unregister = self._content.get('allow_unregister', True)
             self._allow_preview = self._content.get('allow_preview', False)
             self._is_lti = self._content.get('is_lti', False)
+            self._lti_url = self._content.get('lti_url', '')
             self._lti_keys = self._content.get('lti_keys', {})
             self._lti_send_back_grade = self._content.get('lti_send_back_grade', False)
             self._tags = {key: Tag(key, tag_dict, self.gettext) for key, tag_dict in self._content.get("tags", {}).items()}
+            task_dispenser_class = task_dispensers.get(self._content.get('task_dispenser', 'toc'), TableOfContents)
+            # Here we use a lambda to encourage the task dispenser to pass by the task_factory to fetch course tasks
+            # to avoid them to be cached along with the course object. Passing the task factory as argument
+            # would require to pass the course too, and have a useless reference back.
+            self._task_dispenser = task_dispenser_class(lambda: self._task_factory.get_all_tasks(self), self._content.get("dispenser_data", ''))
         except:
             raise Exception("Course has an invalid YAML spec: " + self.get_id())
 
@@ -59,12 +95,37 @@ class WebAppCourse(Course):
             self._registration_ac = None
             self._registration_ac_list = []
             self._groups_student_choice = False
-            self._use_classrooms = True
             self._allow_unregister = False
         else:
             self._lti_keys = {}
+            self._lti_url = ''
             self._lti_send_back_grade = False
 
+        # Build the regex for the ACL, allowing for fast matching. Only used internally.
+        self._registration_ac_regex = self._build_ac_regex(self._registration_ac_list)
+
+    def get_translation_obj(self, language):
+        return self._translations.get(language, gettext.NullTranslations())
+
+    def gettext(self, language, *args, **kwargs):
+        return self.get_translation_obj(language).gettext(*args, **kwargs)
+
+    def get_id(self):
+        """ Return the _id of this course """
+        return self._id
+
+    def get_fs(self):
+        """ Returns a FileSystemProvider which points to the folder of this course """
+        return self._fs
+
+    def get_task(self, taskid):
+        """ Returns a Task object """
+        return self._task_factory.get_task(self, taskid)
+
+    def get_descriptor(self):
+        """ Get (a copy) the description of the course """
+        return copy.deepcopy(self._content)
+
     def get_staff(self):
         """ Returns a list containing the usernames of all the staff users """
         return list(set(self.get_tutors() + self.get_admins()))
@@ -81,7 +142,7 @@ class WebAppCourse(Course):
         """ Returns true if the course is accessible by users that are not administrator of this course """
         return self.get_accessibility().is_open()
 
-    def is_registration_possible(self, user_info):
+    def is_registration_possible(self, user_info: UserInfo):
         """ Returns true if users can register for this course """
         return self.get_accessibility().is_open() and self._registration.is_open() and self.is_user_accepted_by_access_control(user_info)
 
@@ -95,32 +156,28 @@ class WebAppCourse(Course):
 
     def get_accessibility(self, plugin_override=True):
         """ Return the AccessibleTime object associated with the accessibility of this course """
-        vals = self._hook_manager.call_hook('course_accessibility', course=self, default=self._accessible)
+        vals = self._plugin_manager.call_hook('course_accessibility', course=self, default=self._accessible)
         return vals[0] if len(vals) and plugin_override else self._accessible
 
     def get_registration_accessibility(self):
         """ Return the AccessibleTime object associated with the registration """
         return self._registration
 
-    def get_tasks(self):
-        return OrderedDict(sorted(list(Course.get_tasks(self).items()), key=lambda t: (t[1].get_order(), t[1].get_id())))
+    def get_tasks(self, ordered=False):
+        return self._task_dispenser.get_ordered_tasks() if ordered else self._task_factory.get_all_tasks(self)
 
     def get_access_control_method(self):
         """ Returns either None, "username", "binding", or "email", depending on the method used to verify that users can register to the course """
         return self._registration_ac
 
-    def get_access_control_list(self):
-        """ Returns the list of all users allowed by the AC list """
+    def get_access_control_list(self) -> List[str]:
+        """ Returns the list of all users/emails/binding methods/... (see get_access_control_method) allowed by the AC list """
         return self._registration_ac_list
 
     def can_students_choose_group(self):
         """ Returns True if the students can choose their groups """
         return self._groups_student_choice
 
-    def use_classrooms(self):
-        """ Returns True if classrooms are used """
-        return self._use_classrooms
-
     def is_lti(self):
         """ True if the current course is in LTI mode """
         return self._is_lti
@@ -129,30 +186,38 @@ class WebAppCourse(Course):
         """ {name: key} for the LTI customers """
         return self._lti_keys if self._is_lti else {}
 
+    def lti_url(self):
+        """ Returns the URL to the external platform the course is hosted on """
+        return self._lti_url
+
     def lti_send_back_grade(self):
         """ True if the current course should send back grade to the LTI Tool Consumer """
         return self._is_lti and self._lti_send_back_grade
 
-    def is_user_accepted_by_access_control(self, user_info):
+    def is_user_accepted_by_access_control(self, user_info: UserInfo):
         """ Returns True if the user is allowed by the ACL """
         if self.get_access_control_method() is None:
             return True
-        elif not user_info:
+
+        keys_per_access_control_method = {
+            "username": (lambda: [user_info.username]),
+            "email": (lambda: [user_info.email]),
+            "binding": (lambda: user_info.bindings.keys())
+        }
+
+        if not user_info or self.get_access_control_method() not in keys_per_access_control_method:
             return False
-        elif self.get_access_control_method() == "username":
-            return user_info["username"] in self.get_access_control_list()
-        elif self.get_access_control_method() == "email":
-            return user_info["email"] in self.get_access_control_list()
-        elif self.get_access_control_method() == "binding":
-            return set(user_info["bindings"].keys()).intersection(set(self.get_access_control_list()))
-        return False
+
+        # check that at least one key matches in the list
+        keys = keys_per_access_control_method[self.get_access_control_method()]()
+        return any(self._registration_ac_regex.match(key) for key in keys)
 
     def allow_preview(self):
         return self._allow_preview
 
     def allow_unregister(self, plugin_override=True):
         """ Returns True if students can unregister from course """
-        vals = self._hook_manager.call_hook('course_allow_unregister', course=self, default=self._allow_unregister)
+        vals = self._plugin_manager.call_hook('course_allow_unregister', course=self, default=self._allow_unregister)
         return vals[0] if len(vals) and plugin_override else self._allow_unregister
 
     def get_name(self, language):
@@ -166,3 +231,13 @@ class WebAppCourse(Course):
 
     def get_tags(self):
         return self._tags
+
+    def get_task_dispenser(self):
+        """
+       :return: the structure of the course
+       """
+        return self._task_dispenser
+
+    def _build_ac_regex(self, list_ac):
+        """ Build a regex for the AC list, allowing for fast matching. The regex is only used internally """
+        return re.compile('|'.join(re.escape(x).replace("\\*", ".*") for x in list_ac))
